# -*- coding: utf-8 -*-
#
# Copyright (c) 2008 elias.soong@gmail.com
# http://www.elias.cn
# All rights reserved.
#
# 本文件是流程挖掘中对重复任务问题进行预处理的程序。
# 基本原理是对输入日志中的重复任务进行换名处理，
# 从而使不支持重复任务处理的挖掘算法也能够支持重复任务日志的模型挖掘。
# 算法的基本原理是对同名任务进行最近邻自动聚类。
# 
# 2008-02-25, Version 0.1:
# 利用minidom界面和最近邻聚类算法初步实现基本功能。

"""
RenameDuplicateTasks.py by elias.soong@gmail.com
"""
__version__ = "0.1"
import xml.dom.minidom
import difflib
import sys
import psyco
psyco.full()

COMPARE_TYPE = "NoSelf"    # 比较字符串的生成策略，包括："Common","WithNone","NoSelf"
COMPARE_WINDOW = 2    # 比较同名事件的参考窗口大小默认值
            # 也即比较某同名事件时，将使用其前后各COMPARE_WINDOW个事件
NN_DISTANCE = 2   # 最近邻聚类的类间距离阈值默认值，也许设置为2比较合适

def renameDuplicateTasks(inputFile, outputFile, compareWindow = None, nnDistance = None):
    """
    对日志中的重复任务利用最近邻聚类算法处理后重新命名。也即程序主函数。
    """
    # 初始化参数：
    if compareWindow == None:
        compareWindow = COMPARE_WINDOW
    if nnDistance == None:
        nnDistance = NN_DISTANCE
    # 调用各函数，开始计算过程：
    doc = xml.dom.minidom.parse(inputFile) # 读入日志文件
    traces = getSimpleTraces(doc)
    eventDic = getDistinctEventDic(traces, compareWindow)
    distanceMatrixDic = genDistanceMatrix(eventDic, compareWindow)
    printDistanceStatistics(distanceMatrixDic)
    tasksClustersDic = doNNClustering(distanceMatrixDic, nnDistance)
    printClustersStatistics(tasksClustersDic)
    renameDic = getRenameDic(eventDic, tasksClustersDic)
    printRenameStatistics(renameDic)
    saveRename(outputFile, doc, renameDic)
    pass

def getSimpleTraces(xmlTracesLog):
    """
    将dom格式的工作流日志简化为列表形式，仅保留各日志事件的名称和出现顺序。
    其中每一条流程日志实例作为一个列表表示，而所有的流程日志示例再组成一个大列表存储。
    """
    xmlProcesses = xmlTracesLog.getElementsByTagName("ProcessInstance")
    traces = [[]] * (len(xmlProcesses) + 1)    # 这里要加1因为流程日志id从1开始，而Python列表下标从0开始。
    # 但是对单条流程记录来说，其中的第一个事件记录的标号则是0，因为是用列表的append()函数输入的。
    for xmlProcess in xmlProcesses:
        id = int(xmlProcess.getAttribute("id"))    # 当前流程实例的id号
        events = []
        xmlEvents = xmlProcess.getElementsByTagName("WorkflowModelElement")
        for xmlEvent in xmlEvents:
            name = ""
            for node in xmlEvent.childNodes:
                if node.nodeType == node.TEXT_NODE:
                    name += node.data
            events.append(name)
        traces[id] = events
#    print(traces)
    return traces

def getIndexRange(tracesList):
    """
    已废弃：计算processId和eventId下标的范围极限。
    """
    return (len(tracesList), max(map(len, tracesList)))

def getDistinctEventDic(tracesList, compareWindow = COMPARE_WINDOW):
    """
    输入流程日志数组，返回一个字典：
    字典的键值是不重名的事件名称，对应的元素是与键值同名的事件列表。
    列表中的事件包括其原始位置（Tuple(原所在流程实例序号,流程日志中序号)）和compareWindow下的对应比较参考信息。
    """
    dic = {}
    for processId in range(len(tracesList)):
        trace = tracesList[processId]    # 这里的trace和process其实是一个意思，只是日志的XML表示习惯叫法不同而已。
        if trace == []:
            continue
        traceLen = len(trace)
        for eventId in range(traceLen):
            eventName = trace[eventId]
            eventInfo = []
            eventInfo.append((processId, eventId))
            if COMPARE_TYPE == "Common":
                eventInfo.append(trace[max(0, eventId - compareWindow):min(traceLen, eventId + compareWindow + 1)])
            elif COMPARE_TYPE == "WithNone":
                eventInfo.append([None] * (compareWindow - eventId) + trace[max(0, eventId - compareWindow):min(traceLen, eventId + compareWindow + 1)] + [None] * (eventId + compareWindow + 1 - traceLen))
            elif COMPARE_TYPE == "NoSelf":
                eventInfo.append(trace[max(0, eventId - compareWindow) : eventId] + trace[eventId + 1 : min(traceLen, eventId + compareWindow + 1)])
            else:
                eventInfo.append([])
            eventList = dic.get(eventName,[])
            eventList.append(eventInfo)
            dic[eventName] = eventList
    return dic

def genDistanceMatrix(eventDic, compareWindow = COMPARE_TYPE):
    """
    计算各同名事件间的距离，返回为矩阵形式。
    目前实际上返回的是字典，也即对应不同事件名称分别计算的小距离矩阵。
    """
    MAX_DISTANCE = (compareWindow * 2 + 1) * 2
    matrixDic = {}
    for eventName in eventDic.keys():
        eventList = eventDic[eventName]
        eventListLen = len(eventList)
        distMatrix = [[MAX_DISTANCE for col in range(eventListLen)] for row in range(eventListLen)]
        for eventLeftIndex in range(eventListLen):
            for eventRightIndex in range(eventListLen):
                distMatrix[eventLeftIndex][eventRightIndex] = dist(eventList[eventLeftIndex][1], eventList[eventRightIndex][1])
        matrixDic[eventName] = distMatrix
#        break    # 这个语句在正常使用时需要去掉。
    return matrixDic

differ=difflib.Differ()
def dist(eventLeft, eventRight):
    """
    计算两个事件的比较窗口下的特征数据所形成的距离。
    也即基于diff算法比较两个事件的特征数据数组。
    """
    diffList = list(differ.compare(eventLeft, eventRight))    # 主要的效率损失就在这句diff
    diffList = map(lambda x:x[0], diffList)
    return diffList.count('+') + diffList.count('-')
#    return 0    # 用这条语句来代替函数体，可以用来验证是否diff语句是性能瓶颈。

def printDistanceStatistics(distanceMatrixDic):
    """
    辅助分析函数：打印距离矩阵的统计数据，也即显示不同的距离长度分别对应多少对事件。
    实际执行会分别显示不同名字的事件的距离矩阵统计情况。
    """
    print "== Statistics for distance matrix =="
    for key in distanceMatrixDic.keys():
        stat = {}
        matrix = distanceMatrixDic[key]
        # 利用循环方式的实现（似乎效率稍微高一些）：
        for i in range(len(matrix)):
            row = matrix[i]
            for j in range(len(row)):
                d = row[j]
                num = stat.get(d, 0)
                num += 1
                stat[d] = num
        # 新的Python风格的实现（但是更慢，不但多了两句列表结构变换，而且count函数的效率恐怕也不高）：
#        combined = reduce(lambda x,y: x + y, matrix)
#        dists = list(set(combined))
#        for d in dists:
#            stat[d] = combined.count(d)
        print key, stat

def doNNClustering(distanceMatrixDic, threshold = NN_DISTANCE):
    """
    对同名事件进行聚类分析。
    聚类算法采用最近邻方法、所用参数在本文件最开头。出于性能考虑，实际使用的输入数据是每种事件的距离矩阵。
    """
    clustersDic = {}
    for eventName in distanceMatrixDic.keys():
        clusters = []
        distanceMatrix = distanceMatrixDic[eventName]
        for i in range(len(distanceMatrix)):
            if len(clusters) == 0:
                clusters.append([i])
            else:
                def getNearestDist(list):    # 返回当前事件与参数指定聚类的最小距离
                    def getDist(x):    # 返回当前事件与参数给定事件的距离
                        return distanceMatrix[i][x]
                    return min(map(getDist,list))
                distToClusters = map(getNearestDist, clusters)
                nearestDist = min(distToClusters)
                if nearestDist <= threshold:
                    clusterIndex = distToClusters.index(nearestDist)
                    clusters[clusterIndex].append(i)
                else:
                    clusters.append([i])
        # 测试聚类按质心距离合并（验证结果只是在某些情况略有改善，效果不明显）：
#        tag = True
#        while tag:
#            tag = False
#            r = range(len(clusters))
#            for i in r:
#                for j in range(i + 1, len(clusters)):
#                    d = 0
#                    for x in clusters[i]:
#                        for y in clusters[j]:
#                            d += distanceMatrix[x][y]
#                    dMean = d / float(len(clusters[i]) * len(clusters[j]))
#                    print dMean
#                    if dMean <= threshold + 1:
#                        tag = True
#                        tmp = clusters.pop(j)
#                        clusters[i] += tmp
#                        r.pop(-1)
#                        break
        # 测试聚类按质心距离合并结束。
        clustersDic[eventName] = clusters
    return clustersDic

def printClustersStatistics(clustersDic):
    """
    打印对各同名事件的聚类统计情况。
    """
    print "== Statistics for clusters =="
    for key in clustersDic.keys():
        print key, map(len, clustersDic[key])

def getRenameDic(eventDic, clustersDic):
    """
    整理聚类信息，生成以流程Id为键值、存着新事件名称的数组为保存值的重命名字典。
    """
    dic = {}
    for key in eventDic.keys():
        events = eventDic[key]
        clusters = clustersDic.get(key,[])
        clustersNum = len(clusters)
        for clusterId in range(clustersNum):
            if clustersNum <= 1:
                newName = None    # 聚类数量不多余一个时，用None表示不修改事件名字
            else:
                newName = key + str(clusterId)    # 如果多于一个聚类结果，那么所有重名事件将被重命名为带从0开始的数字后缀
            for eventIndex in clusters[clusterId]:
                processId, eventId = events[eventIndex][0]
                newNameList = dic.get(processId, [])
                newNameList += [None] * (eventId + 1 - len(newNameList))    # 为将要写入的下标延伸数组长度
                newNameList[eventId] = newName
                dic[processId] = newNameList
    return dic

def printRenameStatistics(renameDic):
    """
    打印对重命名情况的统计结果。也即不变的和将被分别改为别的名字的事件各有多少。
    """
    print "== Statistics for new names =="
    dic = {}
    for nameList in renameDic.values():
        for name in nameList:
            value = dic.get(name, 0)
            value += 1
            dic[name] = value
    print dic

def saveRename(outputFile, doc, renameDic):
    """
    实施重命名过程，输入为指定的XML文件。
    """
    for node in doc.getElementsByTagName("ProcessInstance"):
        processId = int(node.getAttribute("id"))
        element = node.getElementsByTagName("WorkflowModelElement")
        newNameList = renameDic[processId]
        index = 0
        for node2 in element:
            for node3 in node2.childNodes:
                if node3.nodeType == node3.TEXT_NODE:
                    newName = newNameList[index]
                    if newName != None:
                        node3.data = newName
                    index += 1
    output = open(outputFile, 'w')
    doc.writexml(output)
    output.close()

if __name__ == '__main__':
    """
    通过命令行取得算法参数。
    参数用空格分割分别为：输入文件路径及名称，比较窗口大小，聚类距离阈值。
    """
    FILE_NAME_EXT = "_renamed"
    inputFile, compareWindow, threshold = (sys.argv[1:] + [None] * 3)[:3]
    inputFile = inputFile or "cpnToolsSimulationLog.mxml"
    compareWindow = compareWindow or COMPARE_WINDOW
    compareWindow = int(compareWindow)
    threshold = threshold or NN_DISTANCE
    threshold = int(threshold)
    tmp = inputFile.split('.')
    tmp[-2] = tmp[-2] + FILE_NAME_EXT
    outputFile = '.'.join(tmp)
    print "## File: %s, Window: %d, Threshold: %d ##" % (inputFile, compareWindow, threshold)
    import time
    startTime = time.time()
    renameDuplicateTasks(inputFile, outputFile, compareWindow, threshold)
    endTime = time.time()
    print "== Finished in %f seconds! ==" % (endTime - startTime)